/*
 * dtaz - DTraceTazTool, a disk activity analysis tool.
 *        Solaris 10, DTrace/GTK.
 *
 * $Id: dtaz.c 32 2006-05-31 15:28:13Z brendan $
 *
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 *
 * COPYRIGHT: Copyright (c) 2006 Brendan Gregg.
 *
 * 29-Jan-2006  Brendan Gregg   Created this
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <dirent.h>
#include <strings.h>
#include <gtk/gtk.h>

#include "dtaz.h"
#include "colours.h"
#include "dialog.h"
#include "disks.h"
#include "dtrace.h"
#include "menu.h"
#include "plot.h"
#include "statusbar.h"
#include "system.h"

/*
 * GTK Widgets
 *
 * all globals have a "dt_" prefix
 */
GtkWidget	*dt_main_window;	/* main window widget */
GtkWidget	*dt_main_table;		/* main table widget */
GtkWidget	*dt_plot_da;		/* plot drawing area */
GtkWidget	*dt_plot_table;		/* plot table for labels */
GtkWidget	*dt_plot_label[3];	/* plot labels */
GtkWidget	*dt_seek_da;		/* seek drawing area */
GtkWidget	*dt_seek_table;		/* seek table for labels */
GtkWidget	*dt_seek_label[3];	/* seek labels */
GtkWidget	*dt_main_menu;		/* main menu */
GtkWidget	*dt_status;		/* status bar */
GtkWidget	*dt_top_table;		/* top table */
GtkWidget	*dt_options_info;	/* options info label */
GtkWidget	*dt_selector;		/* disk selector */
GtkWidget	*dt_start_b;		/* start tracing button */
GtkWidget	*dt_stop_b;		/* stop tracing button */

/*
 * Globals, with initial values
 */
int		dt_resolution = 4;	/* pixel resolution */
uint64_t	dt_time = 0;		/* current time (ms) */
int		dt_current_x = 0;	/* current x loc counter */
int		dt_seek_last_ym = 0;	/* current seek "max" y pxl */
int		dt_seek_last_ya = 0;	/* current seek "avg" y pxl */
int64_t		dt_seek_last_raw = 0;	/* last raw seek loc */
int64_t		dt_seek_max_raw = 0;	/* max raw seek loc */
int64_t		dt_seek_sum_raw = 0;	/* sum of raw seek locs */
int		dt_seek_count = 0;	/* number of seeks */
int		dt_delay = 100;		/* timeout delay ms */
int		dt_update = 500;	/* screen update ms */
int		dt_update_now = 0;	/* screen update counter */
int		dt_colour_factor = 16;	/* scaling for the colours */
int		dt_debug = 0;		/* debug messages */
int		dt_plot_resized = TRUE;	/* plot resized */
int		dt_plot_exposed = FALSE;	/* plot exposed */

/* plot and seek canvas */
uint16_t	dt_plot_canvas[DT_PLOT_MAX_X][DT_PLOT_MAX_Y];
dt_seek_entry_t	dt_seek_canvas[DT_PLOT_MAX_X];

/* maximum plot y RAW */
int64_t		dt_plot_max_raw = DT_PLOT_MAX_DEF;

/* plot buckets */
int		dt_plot_buckets[DT_PLOT_MAX_Y];

/* scale state */
enum dt_scale	dt_scale = DT_SCALE_INITIAL;

/* private variables */
static int	dt_exposed_count = 0;		/* redraw hysteresis */

/* private functions */
static gint	process_data(gpointer);
static void	update_screen(void);
static void	plot_exposed(void);
static void	redraw_plot(void);
static void	redraw_seek(void);

static gint
enable_delete(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	reset_stdin(0);
	gtk_main_quit();
	return (TRUE);
}

void
start_tracing()
{
	/* check if tracing already */
	if (dt_state == DT_STATE_STARTED) {
		sb_err(dt_status, "already tracing.");
		return;
	}

	if (disks_set_edev(dt_selector) != TRUE)
		return;

	/* start dtrace */
	sb_msg(dt_status, "Starting DTrace...");
	if (dtrace_start(DT_SCRIPT_SINGLE, disks_edev) == FALSE) {
		sb_err(dt_status, "Starting DTrace failed.");
		return;
	}

	/* begin time  */
	dt_time = 0;
	dt_update_now = 0;

	/* set seek y locs to effective 0 */
	dt_seek_last_ya = dt_seek_da->allocation.height;
	dt_seek_last_ym = dt_seek_da->allocation.height;
}

void
stop_tracing()
{
	/* check if tracing first */
	if (dt_state != DT_STATE_STARTED) {
		sb_err(dt_status, "not tracing.");
		return;
	}

	/* terminate dtrace */
	if (dtrace_terminate() == FALSE)
		sb_err(dt_status, "problem closing dtrace");

	/* update statusbar */
	sb_msg(dt_status, "Stopped DTrace.");
}

gint
process_data(gpointer data)
{
	int		index;		/* plot canvas index */
	int 		update;		/* flag to update screen */
	int		seek;
	int		num;
	char		buf[BUF_SIZE];
	long long	block;		/* disk block address */

	/* redraw if needed */
	if (dt_plot_exposed == TRUE && ++dt_exposed_count > 250 / dt_delay) {
		redraw_plot();
		redraw_seek();
		dt_plot_exposed = FALSE;
		dt_exposed_count = 0;
	}

	/* skip rest if not tracing */
	if (dt_state != DT_STATE_STARTED) {
		return (TRUE);
	}

	/* check dtrace pulse */
	if (dtrace_checkpulse(dt_delay) == FALSE) {

		/* dtrace abruptly stopped */
		sb_msg(dt_status, "WARNING: DTrace stopped.");
		dialog_message(dt_main_window, "WARNING",
		    dialog_stopped_msg);

		return (TRUE);
	}

	/*
	 * Read DTrace output
	 */
	update = 0;
	while (1) {

		/* read output */
		num = dtrace_read(buf);
		if (!num)
			break;

		/* heartbeat */
		if (buf[0] == 'h') {
			/* update timer */
			dt_time += dt_heartbeat;
			dt_update_now += dt_heartbeat;

			/* time to update the screen */
			if (dt_update_now >= dt_update) {
				update_screen();
				dt_update_now -= dt_update;
			}

			continue;
		}

		if (buf[0] == '\n')
			continue;

		/* data intput from dtrace */
		block = 0;
		sscanf(buf, "%llx", &block);

		if (block > dt_plot_max_raw && dt_debug)
			g_print("WARNING: block %llu > expected max\n", block);

		/* update plot canvas */
		index = plot_get_index(block);
		dt_plot_canvas[dt_current_x][index]++;	/* IOPS based */

		/* update seek counters */
		seek = dt_seek_last_raw - block;
		seek = seek < 0 ? - seek : seek;
		if (seek < 0)
			seek = -seek;
		dt_seek_sum_raw += seek;
		if (seek > dt_seek_max_raw)
			dt_seek_max_raw = seek;
		dt_seek_count++;
		dt_seek_last_raw = block;
	}

	return (TRUE);
}

void
update_screen(void)
{
	int	plot_max_x = dt_plot_da->allocation.width;
	int	plot_max_y = dt_plot_da->allocation.height;
	int	seek_max_y = dt_seek_da->allocation.height;
	int	seek_avg;
	int	seek_new_ym;
	int	seek_new_ya;
	int	scale;			/* colour scale */
	char	*msg;
	int	i;
	int	j;
	int		bucket;

	/*
	 * Update block plot
	 */

	/* clear block plot ahead */
	gdk_gc_set_foreground(c_gc, &c_white);
	gdk_draw_rectangle(dt_plot_da->window, c_gc, TRUE,
	    dt_current_x * dt_resolution + 5 * dt_resolution, 0,
	    dt_resolution, plot_max_y);

	/* draw grey marker */
	gdk_gc_set_foreground(c_gc, &c_grey);
	gdk_draw_rectangle(dt_plot_da->window, c_gc, TRUE,
	    dt_current_x * dt_resolution + 1 * dt_resolution, 0,
	    1, plot_max_y);

	/* rescale canvas to our max_y */
	for (i = 0; i < DT_PLOT_MAX_Y; i++) {
		/* rescale */
		bucket = plot_get_bucket(i, plot_max_y);
		dt_plot_buckets[bucket] += dt_plot_canvas[dt_current_x][i];

		/* clear plot canvas ahead */
		dt_plot_canvas[dt_current_x + 5][i] = 0;
	}

	/* draw coloured blocks */
	for (i = 0; i <= (plot_max_y / dt_resolution); i++) {

		/* set line colour */
		if (dt_plot_buckets[i] == 0)
			gdk_gc_set_foreground(c_gc, &c_white);
		else {
			scale = dt_plot_buckets[i] / dt_colour_factor;
			if (scale >= c_colours_max)
				scale = c_colours_max - 1;
			gdk_gc_set_foreground(c_gc, &c_colours[scale]);
		}

		/* draw box */
		gdk_draw_rectangle(dt_plot_da->window, c_gc, TRUE,
		    dt_current_x * dt_resolution, i * dt_resolution,
		    dt_resolution, dt_resolution);

		/* reset bucket */
		dt_plot_buckets[i] = 0;
	}

	/*
	 * Update seek plot
	 */

	/* calculate avg and y locs */
	seek_avg = dt_seek_count > 0 ? dt_seek_sum_raw / dt_seek_count : 0;
	seek_new_ya = plot_get_seek_loc(seek_avg, seek_max_y);
	seek_new_ym = plot_get_seek_loc(dt_seek_max_raw, seek_max_y);

	/* save canvas values */
	dt_seek_canvas[dt_current_x].avg = seek_avg > 0 ? seek_avg : 0;
	dt_seek_canvas[dt_current_x].max = dt_seek_max_raw > 0 ?
	    dt_seek_max_raw : 0;

	/* clear seek plot ahead */
	gdk_gc_set_foreground(c_gc, &c_white);
	gdk_draw_rectangle(dt_seek_da->window, c_gc, TRUE,
	    (dt_current_x * dt_resolution) + (5 * dt_resolution), 0,
	    dt_resolution, seek_max_y);

	/* draw max line */
	gdk_gc_set_foreground(c_gc, &c_red);
	gdk_draw_line(dt_seek_da->window, c_gc, dt_current_x * dt_resolution,
	    dt_seek_last_ym, dt_current_x * dt_resolution + dt_resolution,
	    seek_new_ym);

	/* draw avg line */
	gdk_gc_set_foreground(c_gc, &c_blue);
	gdk_draw_line(dt_seek_da->window, c_gc, dt_current_x * dt_resolution,
	    dt_seek_last_ya, dt_current_x * dt_resolution + dt_resolution,
	    seek_new_ya);

	/* save values */
	dt_seek_last_ya = seek_new_ya;
	dt_seek_last_ym = seek_new_ym;
	dt_seek_max_raw = 0;
	dt_seek_sum_raw = 0;
	dt_seek_count = 0;

	/* update x loc counter */
	dt_current_x++;
	if (dt_current_x > plot_max_x / dt_resolution) {

		/* clear a stripe at the start of each plot */
		gdk_gc_set_foreground(c_gc, &c_white);
		gdk_draw_rectangle(dt_plot_da->window, c_gc, TRUE,
		    0, 0, dt_resolution * 6, plot_max_y);
		gdk_draw_rectangle(dt_seek_da->window, c_gc, TRUE,
		    0, 0, dt_resolution * 6, seek_max_y);

		/* clear a stripe on the canvas */
		for (j = 0; j < 6; j++)
			for (i = 0; i < DT_PLOT_MAX_Y; i++)
				dt_plot_canvas[j][i] = 0;

		/* reset x */
		dt_current_x = 0;
	}

	/* update statusbar */
	msg = g_strdup_printf("time %llu ms", (u_longlong_t)dt_time);
	sb_msg(dt_status, msg);
	g_free(msg);
}

void
redraw_plot(void)
{
	int	plot_max_x = dt_plot_da->allocation.width;
	int	plot_max_y = dt_plot_da->allocation.height;
	int	plot_buckets[1024];
	int	i;
	int	j;
	int	bucket;
	int	scale;

	/* redraw plot */
	for (j = 0; j <= plot_max_x / dt_resolution; j++) {

		/* saftey */
		if (j > DT_PLOT_MAX_X)
			break;

		for (i = 0; i < DT_PLOT_MAX_Y; i++)
			plot_buckets[i] = 0;

		/* rescale canvas to our max_y */
		for (i = 0; i < DT_PLOT_MAX_Y; i++) {
			bucket = plot_get_bucket(i, plot_max_y);
			plot_buckets[bucket] += dt_plot_canvas[j][i];
		}

		/* draw coloured blocks */
		for (i = 0; i < (plot_max_y / dt_resolution); i++) {

			/* set line colour */
			if (plot_buckets[i] == 0)
				gdk_gc_set_foreground(c_gc, &c_white);
			else {
				scale = plot_buckets[i] / dt_colour_factor;
				if (scale >= c_colours_max)
					scale = c_colours_max - 1;
				gdk_gc_set_foreground(c_gc, &c_colours[scale]);
			}

			/* draw box */
			gdk_draw_rectangle(dt_plot_da->window, c_gc, TRUE,
			    j * dt_resolution, i * dt_resolution,
			    dt_resolution, dt_resolution);

			/* reset bucket */
			plot_buckets[i] = 0;
		}
	}

	/* draw grey marker */
	gdk_gc_set_foreground(c_gc, &c_grey);
	gdk_draw_rectangle(dt_plot_da->window, c_gc, TRUE,
	    dt_current_x * dt_resolution + 1 * dt_resolution, 0,
	    1, plot_max_y);
}

void
redraw_seek(void)
{
	int	seek_max_x = dt_seek_da->allocation.width;
	int	seek_max_y = dt_seek_da->allocation.height;
	int	seek_new_ya;
	int	seek_new_ym;
	int	seek_last_ya;
	int	seek_last_ym;
	int	j;
	int	last_x;

	last_x = dt_current_x > 0 ? dt_current_x - 1 : 0;
	seek_last_ya = seek_max_y;
	seek_last_ym = seek_max_y;

	/* clear entire seek plot */
	gdk_gc_set_foreground(c_gc, &c_white);
	gdk_draw_rectangle(dt_seek_da->window, c_gc, TRUE,
	    0, 0, seek_max_x, seek_max_y);

	/* redraw plot */
	for (j = 0; j <= seek_max_x / dt_resolution; j++) {

		/* saftey */
		if (j > DT_PLOT_MAX_X)
			break;

		/* skip clear zone */
		if (j > last_x && j < last_x + 6) {
			seek_last_ya = seek_max_y;
			seek_last_ym = seek_max_y;
			continue;
		}

		/* calculate avg and y locs */
		seek_new_ya = plot_get_seek_loc(dt_seek_canvas[j].avg,
		    seek_max_y);
		seek_new_ym = plot_get_seek_loc(dt_seek_canvas[j].max,
		    seek_max_y);

		/* draw max line */
		gdk_gc_set_foreground(c_gc, &c_red);
		gdk_draw_line(dt_seek_da->window, c_gc, j * dt_resolution,
		    seek_last_ym, j * dt_resolution + dt_resolution,
		    seek_new_ym);

		/* draw avg line */
		gdk_gc_set_foreground(c_gc, &c_blue);
		gdk_draw_line(dt_seek_da->window, c_gc, j * dt_resolution,
		    seek_last_ya, j * dt_resolution + dt_resolution,
		    seek_new_ya);

		/* remember last pos */
		seek_last_ya = seek_new_ya;
		seek_last_ym = seek_new_ym;
	}
}

void
update_options_info(void)
{
	char	buf[LINE_LEN];

	/* set new label */
	sprintf(buf, "%dms (CF:%d)", dt_update, dt_colour_factor);
	gtk_label_set_text(GTK_LABEL(dt_options_info), buf);
}

void
plot_exposed(void)
{
	int	plot_max_y = dt_plot_da->allocation.height;
	static int	last_max_y = 0;

	/* plot was (probably) exposed */
	dt_plot_exposed = TRUE;
	dt_exposed_count = 0;

	/* check we really did resize */
	if (plot_max_y != last_max_y)
		dt_plot_resized = TRUE;

	last_max_y = plot_max_y;
}

int
main(int argc, char **argv) {

	gchar	buf[LINE_LEN];

	gtk_init(&argc, &argv);

	/* create main window */
	dt_main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(dt_main_window), "DTraceTazTool");
	g_signal_connect(GTK_OBJECT(dt_main_window), "delete_event",
	    GTK_SIGNAL_FUNC(enable_delete), NULL);

	/* create main table */
	dt_main_table = gtk_table_new(2, 5, FALSE);
	gtk_container_add(GTK_CONTAINER(dt_main_window), dt_main_table);

	/* create main menu */
	menu_create_main(dt_main_window, &dt_main_menu);
	gtk_table_attach(GTK_TABLE(dt_main_table), dt_main_menu,
	    0, 2, 0, 1, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);

	/* create disk selector */
	dt_top_table = gtk_table_new(4, 1, FALSE);
	gtk_table_attach(GTK_TABLE(dt_main_table), dt_top_table,
	    1, 2, 1, 2, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);
	dt_selector = gtk_combo_new();
	disks_find(dt_selector, "/dev/dsk");
	gtk_table_attach(GTK_TABLE(dt_top_table), dt_selector,
	    0, 1, 0, 1, GTK_SHRINK, GTK_SHRINK, 10, 0);
	g_signal_connect(G_OBJECT(GTK_COMBO(dt_selector)->entry), "changed",
	    GTK_SIGNAL_FUNC(disks_combo_select), NULL);

	/* create button */
	dt_start_b = gtk_button_new_with_label("Start");
	gtk_signal_connect(GTK_OBJECT(dt_start_b), "clicked",
	    GTK_SIGNAL_FUNC(start_tracing), NULL);
	gtk_table_attach(GTK_TABLE(dt_top_table), dt_start_b,
	    1, 2, 0, 1, GTK_SHRINK, GTK_SHRINK, 2, 0);
	dt_stop_b = gtk_button_new_with_label("Stop");
	gtk_signal_connect(GTK_OBJECT(dt_stop_b), "clicked",
	    GTK_SIGNAL_FUNC(stop_tracing), NULL);
	gtk_table_attach(GTK_TABLE(dt_top_table), dt_stop_b,
	    2, 3, 0, 1, GTK_SHRINK, GTK_SHRINK, 2, 0);

	/* create options status */
	dt_options_info = gtk_label_new("");
	gtk_table_attach(GTK_TABLE(dt_top_table), dt_options_info,
	    3, 4, 0, 1, GTK_SHRINK, GTK_SHRINK, 10, 0);

	/* create plot drawing area */
	dt_plot_da = gtk_drawing_area_new();
	gtk_widget_set_size_request(dt_plot_da, DT_PLOT_INI_X, DT_PLOT_INI_Y);
	gtk_table_attach(GTK_TABLE(dt_main_table), dt_plot_da,
	    1, 2, 2, 3, GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 5, 2);
	g_signal_connect(GTK_OBJECT(dt_plot_da), "expose_event",
	    GTK_SIGNAL_FUNC(plot_exposed), NULL);

	/* create plot label table and labels */
	dt_plot_table = gtk_table_new(1, 3, FALSE);
	gtk_container_set_border_width(GTK_CONTAINER(dt_plot_table), 5);
	gtk_table_attach(GTK_TABLE(dt_main_table), dt_plot_table,
	    0, 1, 2, 3, GTK_SHRINK, GTK_EXPAND | GTK_FILL, 2, 2);
	dt_plot_label[0] = gtk_label_new("Block 0");
	gtk_table_attach(GTK_TABLE(dt_plot_table), dt_plot_label[0],
	    0, 1, 0, 1, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);
	dt_plot_label[1] = gtk_label_new(NULL);
	gtk_table_attach(GTK_TABLE(dt_plot_table), dt_plot_label[1],
	    0, 1, 1, 2, GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
	sprintf(buf, "Block\n%llu", (u_longlong_t)dt_plot_max_raw);
	dt_plot_label[2] = gtk_label_new(buf);
	gtk_table_attach(GTK_TABLE(dt_plot_table), dt_plot_label[2],
	    0, 1, 2, 3, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);

	/* create seek drawing area */
	dt_seek_da = gtk_drawing_area_new();
	gtk_widget_set_size_request(dt_seek_da, DT_SEEK_INI_X, DT_SEEK_INI_Y);
	gtk_table_attach(GTK_TABLE(dt_main_table), dt_seek_da,
	    1, 2, 3, 4, GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 5, 2);

	/* create seek label table and labels */
	dt_seek_table = gtk_table_new(1, 3, FALSE);
	gtk_container_set_border_width(GTK_CONTAINER(dt_seek_table), 5);
	gtk_table_attach(GTK_TABLE(dt_main_table), dt_seek_table,
	    0, 1, 3, 4, GTK_SHRINK, GTK_EXPAND | GTK_FILL, 2, 2);
	dt_seek_label[0] = gtk_label_new(NULL);
	gtk_label_set_markup(GTK_LABEL(dt_seek_label[0]),
	    "<span foreground=\"red\">(MAX)</span>");
	gtk_table_attach(GTK_TABLE(dt_seek_table), dt_seek_label[0],
	    0, 1, 0, 1, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);
	dt_seek_label[1] = gtk_label_new(NULL);
	gtk_label_set_markup(GTK_LABEL(dt_seek_label[1]),
	    "<span foreground=\"blue\">(AVG)</span>");
	gtk_table_attach(GTK_TABLE(dt_seek_table), dt_seek_label[1],
	    0, 1, 1, 2, GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
	dt_seek_label[2] = gtk_label_new("  Seek\nDistance");
	gtk_table_attach(GTK_TABLE(dt_seek_table), dt_seek_label[2],
	    0, 1, 2, 3, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);

	/* create statusbar */
	dt_status = gtk_statusbar_new();
	gtk_table_attach(GTK_TABLE(dt_main_table), dt_status,
	    0, 2, 4, 5, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);
	sb_msg(dt_status, "Select a disk.");

	/* activate */
	gtk_widget_show_all(dt_main_window);

	/* set drawing areas background colour */
	colours_set(dt_main_window, TRUE);
	gdk_window_set_background(dt_plot_da->window, &c_white);
	gdk_window_set_background(dt_seek_da->window, &c_white);

	/* create timeout */
	g_timeout_add(dt_delay, process_data, NULL);

	/* set initial options info */
	update_options_info();

	/* non GTK init */
	disable_zombies();
	(void) sigignore(SIGTTOU);

	/* Go! */
	gtk_main();

	return (0);
}
