/*
 * system.c - system functions.
 *
 * $Id: system.c 12 2006-02-05 06:51:33Z brendan $
 *
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 *
 * COPYRIGHT: Copyright (c) 2006 Brendan Gregg.
 *
 * 29-Jan-2006  Brendan Gregg   Created this
 */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <signal.h>
#include <unistd.h>
#include <termios.h>
#include <sys/filio.h>

#include "system.h"

void
raw_stdin(void)
{
	int		opt;
	struct termios  term;

	/* undo changes on signal */
	(void) signal(SIGINT, reset_stdin);
	(void) signal(SIGBUS, reset_stdin);
	(void) signal(SIGSEGV, reset_stdin);
	(void) signal(SIGTSTP, reset_stdin);
	if (tcgetattr(STDIN_FILENO, &term) == -1)
		die_err("getting stdin attrs.");

	/* switch off canonical mode */
	term.c_cc[VMIN] = 1;
	term.c_lflag &= ~(ICANON | IEXTEN);

	if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &term) == -1)
		die_err("writing stdin attrs.");

	/* poke stdin */
	opt = 1;
	if (ioctl(STDIN_FILENO, FIONBIO, &opt) == -1)
		msg_err("ioctl to set STDIN to raw mode failed");
}

/* ARGSUSED */
void
reset_stdin(int sig)
{
	/*
	 * if the program terminates while non-blocking is still on,
	 * your shell immediately exits (it's expecting blocking reads).
	 * this is annoying, so we put blocking back to normal on exit.
	 */
	int opt = 0;
	if (ioctl(STDIN_FILENO, FIONBIO, &opt) == -1)
		msg_err("ioctl to reset STDIN failed");
}

void
die_blocking(void)
{
	reset_stdin(0);
	exit(1);
}

void
die_err(const char *msg)
{
	reset_stdin(0);
	(void) fprintf(stderr, "ERROR: %s\n", msg);
	exit(1);
}

void
msg_err(const char *msg)
{
	/* this is a function so that all warnings can be tweaked */
	(void) fprintf(stderr, "WARNING: %s\n", msg);
}

void
disable_zombies(void)
{
	struct sigaction	act;
	struct sigaction	oact;

	/*
	 * disable zombies
	 *
	 * BDG - I'd rather read child status for improved
	 * error messages, but GTK seems to interfere with waitpid.
	 */
	act.sa_handler = SIG_IGN;
	act.sa_flags = SA_NOCLDWAIT;
	if (sigaction(SIGCLD, &act, &oact) == -1)
		msg_err("disabling zombies failed");
}
